/**
 *  @file bleInterface.cpp
 *  @brief Provides data transmission over BLE (encapsulating BLE details)
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */

// BLE Interface for Adafruit BLE
#include "bleInterface.h"

#include "debug.h"

// Define to select the UUID model to be used
#define USE_HM10_UUID

/**
 * Callback invoked when central connects
 */
void BleInterface::cbConnect(uint16_t conn_handle) {

   // Get reference to current connection
   BLEConnection *conn = Bluefruit.Connection(conn_handle);

   // see ble_gap.h:
   // #define BLE_GAP_PHY_AUTO     0x00    /**< Automatic PHY selection. Refer @ref sd_ble_gap_phy_update for more information.*/
   // #define BLE_GAP_PHY_1MBPS    0x01    /**< 1 Mbps PHY. */
   // #define BLE_GAP_PHY_2MBPS    0x02    /**< 2 Mbps PHY. */
   // #define BLE_GAP_PHY_CODED    0x04    /**< Coded PHY. */
   // #define BLE_GAP_PHY_NOT_SET  0xFF    /**< PHY is not configured. */

   // Request PHY 1 Mbit
   conn->requestPHY(BLE_GAP_PHY_1MBPS);
   // Request to update data length
   conn->requestDataLengthUpdate();
   // Request MTU size
   conn->requestMtuExchange(MtuSize);

   // Request shorter connection interval (in unit of 1.25 ms)
   // (needed for constant data flow)
   // bool requestConnectionParameter(uint16_t conn_interval, uint16_t slave_latency = BLE_GAP_CONN_SLAVE_LATENCY, uint16_t sup_timeout = BLE_GAP_CONN_SUPERVISION_TIMEOUT_MS/10);
   conn->requestConnectionParameter(12, 0);

   // char central_name[32+1] = { 0 };
   // conn->getPeerName(central_name, sizeof(central_name));
}

/**
 * Callback invoked when central disconnects
 */
void BleInterface::cbDisconnect(uint16_t conn_handle, uint8_t reason) {
   (void) conn_handle;
   (void) reason;
}

/**
 * Should be called periodically to poll BLE UART FIFO
 * (default implementation, does nothing with data)
 */
void BleInterface::poll() {

   uint8_t ch = 0x00;
   while (bleuart.read(& ch, 1)) {
      ; // nothing
   }
}

/**
 * BLE connected query
 * @return 1 if a BLE connection is active, else 0
 */
int BleInterface::isConnected(void) {

   return Bluefruit.connected() > 0 ? 1 : 0;
}

/**
 * Stop BLE advertising and force disconnect
 */
void BleInterface::shutdown(void) {

   Bluefruit.Advertising.restartOnDisconnect(false);
   Bluefruit.Advertising.stop();
   uint16_t hd = Bluefruit.connHandle();
   Bluefruit.disconnect(hd);
}

/* For future extension
void bleuart_rx_callback(uint16_t conn_hdl) {
  (void) conn_hdl;
}

void bleuart_notify_callback(uint16_t conn_hdl, bool enabled) {
  (void) conn_hdl;
  (void) enabled;
}
*/

/**
 * Transmit a string via BLE
 * @param buf  String / Data
 * @param len  Length
 */
void BleInterface::sendData(const char *buf, int len) {

   bleuart.write(buf, len);
}

/**
 * Setup and start BLE advertising
 */
void BleInterface::startAdv(void) {

   Bluefruit.Advertising.addFlags(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
   Bluefruit.Advertising.addTxPower();

   // Add name to BLE 'advertising set'
   // Maximum data length is 31
   Bluefruit.Advertising.addName();
   // Add name also to secondary scan response packet
   Bluefruit.ScanResponse.addName();

   // Add device information service
   // Bluefruit.Advertising.addService(bledis);

   // Add UART service
   Bluefruit.Advertising.addService(bleuart);

   // Enable auto advertising if disconnected
   Bluefruit.Advertising.restartOnDisconnect(true);

   // Set the advertising interval in units of 0.625 ms
   // Fast mode = 20 ms, Slow mode = 152.5 ms (units of 0.625 ms)
   // see https://developer.apple.com/library/content/qa/qa1931/_index.html
   Bluefruit.Advertising.setInterval(BLE_ADV_INTERVAL_FAST_DFLT, BLE_ADV_INTERVAL_SLOW_DFLT);

   // Timeout for fast mode 30 seconds
   Bluefruit.Advertising.setFastTimeout(BLE_ADV_FAST_TIMEOUT_DFLT);

   // Never stop advertising
   Bluefruit.Advertising.start(0);

   // Bluefruit.Advertising.setStopCallback(adv_stop_callback);
   // Bluefruit.Advertising.start(ADV_TIMEOUT);
}

/**
 * Initialise the Adafruit BLE library
 * @param deviceName   Name of the device visible at BLE central
 */
void BleInterface::init(const char *deviceName) {

   // Bluefruit must NOT flash the LED
   Bluefruit.autoConnLed(false);

   // NOTE: All config* function must be called before begin()

   // NOTE: Do not use the bandwith setting function configPrphBandwidth()
   // It hides only the specific settings for configPrphConn

   // BLE Event length: Time set aside for this connection on every connection interval in 1.25 ms units

   // Best with MTU size 23 (9.5/sec)
   Bluefruit.configPrphConn(MtuSize, 3, 4, 1);
   // Best with MTU size 43 (9.5/sec) - fifo can be a bit shorter
   // Bluefruit.configPrphConn(MtuSize, 3, 3, 1);

   Bluefruit.begin();

   // Set max power for nRF52840
   // -40dBm, -20dBm, -16dBm, -12dBm, -8dBm, -4dBm, 0dBm, +2dBm, +3dBm, +4dBm, +5dBm, +6dBm, +7dBm, +8dBm
   Bluefruit.setTxPower(-12);

   // Name used also in advertising if addName() called
   Bluefruit.setName(deviceName);
   // Bluefruit.setName(getMcuUniqueID());

   Bluefruit.Periph.setConnectCallback(cbConnect);
   Bluefruit.Periph.setDisconnectCallback(cbDisconnect);

   // optional: Device Information Service
   // bledis.setManufacturer("Seeed"); // or use MNA
   // bledis.setModel("XIAOnRF"); // or use MMO
   // ...
   // bledis.begin();

   bleuart.begin();
}
