/**
 *  @file bleInterface.h
 *  @brief Provides data transmission over BLE (encapsulating BLE details)
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */
#ifndef _BLEINTERFACE_H
#define _BLEINTERFACE_H

// Custom UART class with Adafruit BLE library
#include "fsUart.h"

// BLE attribute length
// - in 4.0 and 4.1, the maximum ATT Payload in a 'Uncoded Packet' is 20 bytes
// - in 4.2 and 5.0, the new Data Length Extension (DLE) allows the ATT Payload
//   to hold up to 244 bytes
// The ArduinoBLE library hard-sets this *packet* length to 20 bytes,
// equal to the BLE attribute length
#define BLE_ATT_LEN     20

/**
 * Encapsulates BLE connection details
 */
class BleInterface {

public:
   void init(const char *deviceName);
   void startAdv(void);
   void sendData(const char *buf, int len);
   void poll(void);
   void shutdown(void);
   int isConnected(void);

   //! BLE MTU size
   /*! NOTE: The connection interval must be rather short for constant data flow anyway,
    * so there is no benefit in using a larger MTU size for the messages we send. */
   const static int MtuSize = BLE_ATT_LEN + 3;

protected:
   //! Customized UART service
   FsUart bleuart;

   // Device information service not really needed
   // BLEDis bledis;

   static void cbDisconnect(uint16_t conn_handle, uint8_t reason);
   static void cbConnect(uint16_t conn_handle);
};


#endif // _BLEINTERFACE_H
