/**
 *  @file fsSensor.cpp
 *  @brief Sketch main module
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */

// -----------------------------------------------------------------------------
// Library headers
// -----------------------------------------------------------------------------

#include <Arduino.h>            // Added by Arduino in build process

#include <Wire.h>               // For I2C comms
#include <sdpsensor.h>          // Sensirion SDP3x library
#include <SensirionI2CSht4x.h>  // Sensirion SHT4x library
#include <Adafruit_DPS310.h>    // Adafruit DPS310 library

// Local modules
#include "pindefs.h"
#include "sounds.h"
#include "debug.h"
#include "ptvBleInterface.h"
#include "deepSleep.h"
#include "pwrSupply.h"

// -----------------------------------------------------------------------------
// Settings
// -----------------------------------------------------------------------------

//! Manufacturer meta data
#define DEVICE_MANUFACTURER  "Flying Silicon"
//! Device name meta data (also sent in BLE advertising, truncated to 31)
#define DEVICE_NAME          "FS-SP4"
//! Serial number meta data
#define DEVICE_SERIAL        "23-001"

//! Run or stay passive (and charge) when USB is connected
#define ACTIVE_WITH_USB  0

//! Flash blue LED (when BLE connected) also when running on battery
#define FLASH_ALWAYS  1

//! Use the long PTVSOAR message format also for data
#define SEND_LONG  0

//! Enter deep sleep when battery drained (else suspend loop)
#define DEEP_SLEEP  1

// -----------------------------------------------------------------------------

// Caution: buggy/immediate deep sleep can prevent flashing via serial!
//! Delay before auto power off (without BLE connection)
#define AUTO_OFF_MS  3600000

//! Number of meta headers to send on (re)connect
#define META_REPEAT  16

//! Patch the red LED which is going on when BLE device is bonded
#define PATCH_LEDR   digitalWrite(LEDR, HIGH)

// -----------------------------------------------------------------------------
// Constants
// -----------------------------------------------------------------------------

//! Message send interval ms (critical)
#define IntervalMSG   125
//! SDP read interval ms
#define IntervalSDP   IntervalMSG
//! DPS read interval ms
#define IntervalDPS   IntervalMSG
//! SHT read interval ms
#define IntervalSHT   1000
//! Power state update interval ms
#define IntervalPWR   1000
//! Time ms before BLE receive repeats
#define IntervalPOL   IntervalMSG

//! Loop interval ms when disabled
#define LoopPassiveMs 125
//! Loop interval when not connected
#define LoopIdleMs    125
//! Delay ms when no message needs to be sent
#define NopDelayMs    10

//! LED flashing interval when BLE is active but disconnected
#define IdleIntervalLED  1250
//! Time LED is on
#define IdleLightLED  625

//! LED flashing interval when BLE is active and connected
#define IntervalLED  2500
//! Time LED is on
#define LightLED  125


//! Startup melody
static const int Melody[] = {
    NOTE_C3, 16,
    NOTE_G3, 16,
    NOTE_E3, 16,
    NOTE_C3, 32,
    NOTE_G3, -16,
    NOTE_E3, 8,
    MELODY_END
};


// -----------------------------------------------------------------------------
// Global singleton objects
// -----------------------------------------------------------------------------

//! Singleton connection to be used (all BLE stuff is singleton)
static PtvBleInterface Remote;

//! Sensirion SDP31 for Pitot
static SDP3XSensor sdp;

//! Sensirion SHT for temperature / humidity
static SensirionI2CSht4x sht;

//! Infineon DPS310 for barometric pressure
static Adafruit_DPS310 dps;

//! Sensor data to work with
static PtvBleInterface::SensorData sensorData;


// -----------------------------------------------------------------------------
// Global status variables
// -----------------------------------------------------------------------------

//! LED blinking state 0/1
static int isBlueLedOn = 0;
//! Operating mode 0/1
static int isEnabled = 0;
//! Connection state 0/1
static int isConnected = 0;

//! Meta data repetition counter
static unsigned int sentMeta = 0;

//! Current loop time ms
static unsigned long timeMs = 0;

//!< Last time when connected via BLE
static unsigned long lastActiveMs;

static unsigned long lastSHTMs; //!< Last read SHT ambient temperature/humidity
static unsigned long lastSDPMs; //!< Last read SDP differential pressure
static unsigned long lastDPSMs; //!< Last read DPS atmospheric pressure
static unsigned long lastMSGMs; //!< Last time message sent
static unsigned long lastPOLMs; //!< Last BLE receive poll
static unsigned long lastLEDMs; //!< Last LED state change
static unsigned long lastPWRMs; //!< Last read of battery/voltage status


/**
 * Initialise all global status variables
 */
static void resetStatus(void) {

   sensorData.diffPa = 0.0;
   sensorData.absHpa = 1013.25;

   sensorData.batVolt = 0.0;
   sensorData.batPct = 0.0;

   sensorData.tempC = 0.0;
   sensorData.humPct = 0.0;

   isBlueLedOn = 0;
   isConnected = 0;
   isEnabled = 0;

   sentMeta = 0;

   PwrSupply::reset();
}

/**
 * Initialise the function timers
 */
static void resetTimers(void) {

   timeMs = millis();

   // initialise auto off timer
   lastActiveMs = timeMs;

   // let all functions execute at next loop
   lastSHTMs = timeMs - IntervalSHT;
   lastPWRMs = timeMs - IntervalPWR;
   lastSDPMs = timeMs - IntervalSDP;
   lastDPSMs = timeMs - IntervalDPS;
   lastMSGMs = timeMs - IntervalMSG;
   lastPOLMs = timeMs - IntervalPOL;

   lastLEDMs = timeMs; // for clean start of flashing
}

/**
 * Switch RGB LEDs off
 */
static void clearLEDs(void) {

   pinMode(LEDB, OUTPUT); // paranoia
   digitalWrite(LEDB, HIGH);

   pinMode(LEDG, OUTPUT); // paranoia
   digitalWrite(LEDG, HIGH);

   pinMode(LEDR, OUTPUT); // paranoia
   digitalWrite(LEDR, HIGH);
}

/**
 * Periodically read differential pressure from SDP31 sensor
 */
static void readSDP(void) {

   const int ret = sdp.readSample();
   if (ret) {
      LOGL("No data from SDP")
      return;
   }

   // correction of presssure differential wrt altitude/QNH
   sensorData.diffPa = sdp.getDifferentialPressure() * (966.00 / sensorData.absHpa);

   // NOTE: forward also negative values so that data processing
   // by filter around zero can work
   LOG("SDP ")
   LOG(sensorData.diffPa)
   LOGL(" Pa")
}

/**
 * Periodically read atmospheric pressure from DPS sensor
 */
static void readDPS(void) {

   // always read pressure and temperature
   // pressure needs temperature for compensation formula
   if (!dps.temperatureAvailable())
      return; // wait until there's something to read
   if (!dps.pressureAvailable())
      return; // wait until there's something to read

   sensors_event_t temp_event, pressure_event;
   dps.getEvents(&temp_event, &pressure_event);

   sensorData.absHpa = pressure_event.pressure;

   LOG("DPS ")
   LOG(sensorData.absHpa)
   LOGL(" hPa")
}

/**
 * Periodically read temperature and humidity from SHT sensor
 */
static void readSHT(void) {

   float t, h;
   sht.measureHighPrecision(t, h);
   sensorData.tempC = t;
   sensorData.humPct = h;

   LOG("SHT ")
   LOG(sensorData.tempC)
   LOG(" C   ")
   LOG(sensorData.humPct)
   LOGL(" RH")
}


/**
 * Flash the blue LED with the specified timing
 * @param  active  enable (1) or disable (0) blinking
 * @param  intervalMs  duration of a blink period in ms
 * @param  lightMs  time in ms that the LED is on
 */
static void flashBlueLED(int active, unsigned int intervalMs, unsigned int lightMs) {

   if (!isBlueLedOn) {
      // wait for the interval to expire before turning it on
      if (timeMs - lastLEDMs >= intervalMs) {
         lastLEDMs += intervalMs;

         if (active) {
            digitalWrite(LEDB, LOW); // switch on
            isBlueLedOn = 1;
         }
      }
   }
   else {
      // wait for the duration to expire before turning LED off
      if (timeMs - lastLEDMs >= lightMs) {
         lastLEDMs += lightMs;

         digitalWrite(LEDB, HIGH); // switch off
         isBlueLedOn = 0;
      }
   }
}

/**
 * Manage power supply status and related stuff
 */
static void managePower() {

   PwrSupply::updateState();

   // data needed to send
   sensorData.batPct = PwrSupply::getBatPct();
   sensorData.batVolt = PwrSupply::getBatVolt();
   sensorData.isCharging = PwrSupply::isCharging();

   // disable buzzer if running with USB only
   if (PwrSupply::isUsbConnected() || PwrSupply::isBatPowered())
      Sounds::setBuzzer(PIN_BUZZER);
   else
      Sounds::setBuzzer(0);
}

/**
 * Process a control command received from remote
 * @param cmd  String holding the command
 */
static void processCmd(const String &cmd) {
   // nothing so far
   LOGL(cmd)
}

/**
 * Called when device becomes activated
 */
static void enable(void) {

   isEnabled = 1;

   // clear receive buffer
   Remote.clrReceived();
   // start advertising again
   Remote.startAdv();

   // Play melody
   Sounds::playMelody(Melody);
   delay(500);

   // Play battery / power status
   if (PwrSupply::isUsbConnected())
      Sounds::playCharging();
   else
      Sounds::playBattery(sensorData.batPct);
   delay(500);

   resetTimers();
   LOGL("Switched on")
}

/**
 * Called when device becomes deactivated
 */
static void disable(void) {

   isEnabled = 0;
   Sounds::playSwitchOff();
   Remote.shutdown();
   LOGL("Disabled")
}

/**
 * Shut down BLE and suspend loop until next restart
 *  - device can still be charged but will not run until a power cycle occurs
 * Calling suspend() never returns!
 */
static void suspend(void) {

   LOGL("Switched off")
   clearLEDs();

   // try a clean BLE disconnect
   Remote.shutdown();
   delay(1000);

   // test if pin state is kept in sleep mode (important)
   // digitalWrite(LEDR, LOW);

#if DEEP_SLEEP
   enterDeepSleep();
   // should never return
#else
   setLowPowerMode();
   // suspend the RTOS loop task
   // immediately effective, does not return
   suspendLoop();
#endif
   // MUST NOT BE REACHED
   Sounds::playSwitchOff();
}

/**
 * Periodically called when device is active and no BLE central is connected
 */
static void loopIdle(void) {

   // Blink slowly (also when running on battery)
   flashBlueLED(1, IdleIntervalLED, IdleLightLED);

   PATCH_LEDR;
   delay(LoopIdleMs);
}

/**
 * Periodically called when device is active and a BLE central is connected
 */
static void loopConnected(void) {

   if (timeMs >= lastDPSMs + IntervalDPS) {
      lastDPSMs += IntervalDPS;
      readDPS();
   }

   if (timeMs >= lastSHTMs + IntervalSHT) {
      lastSHTMs += IntervalSHT;
      readSHT();
   }

   if (timeMs >= lastSDPMs + IntervalSDP) {
      lastSDPMs += IntervalSDP;
      readSDP();
   }

   if (timeMs >= lastMSGMs + IntervalMSG) {
      lastMSGMs += IntervalMSG;

      if (sentMeta < META_REPEAT) {
         Remote.sendMetaMsg(DEVICE_MANUFACTURER, DEVICE_NAME, DEVICE_SERIAL);
         sentMeta ++;
      }
      else if (SEND_LONG) {
         Remote.sendLongMsg(sensorData);
      }
      else {
         Remote.sendShortMsg(sensorData);
      }

      PATCH_LEDR;
   }
   else {
      // if no message sent, sleep some ms
      delay(NopDelayMs);
   }

   if (timeMs >= lastPOLMs + IntervalPOL) {
      lastPOLMs += IntervalPOL;
      String cmd;
      if (Remote.getReceived(cmd)) {
         processCmd(cmd);
      }
   }

   // Short flashes when BLE connection established
   flashBlueLED(PwrSupply::isUsbConnected() || FLASH_ALWAYS, IntervalLED, LightLED);
}

/**
 * Periodically called when device is inactive
 */
static void loopPassive(void) {

   PATCH_LEDR;
   // constant LED on when inactive and charging
   digitalWrite(LEDB, sensorData.isCharging ? LOW : HIGH);

   delay(LoopPassiveMs);
}


/**
 * Application loop
 */
void appLoop(void) {

   timeMs = millis();

   if (timeMs >= lastPWRMs + IntervalPWR) {
      lastPWRMs += IntervalPWR;
      managePower();
   }

   if (PwrSupply::isBatDrained()) {
      // silent power off
      suspend();
      return;
   }

   if (timeMs >= lastActiveMs + AUTO_OFF_MS) {
      // power off with sound
      Sounds::playSwitchOff();
      suspend();
      return;
   }

   if (DEBUG || ACTIVE_WITH_USB || PwrSupply::isBatPowered()) {

      // battery connected or run with USB set
      if (! isEnabled)
         enable();

      if (! Remote.isConnected()) {
         // idle, no sensor data needed

         // play disconnect sound when connected before
         if (isConnected > 0) {
            LOGL("Disconnected")
            Sounds::playDisconnect();
         }
         isConnected = 0;
         sentMeta = 0;

         // prevent auto off if running on USB
         // if (! isBattPowered())
         //   lastActiveMs = timeMs;
         loopIdle();
      }
      else {
         // full ooperation
         // NOTE: we get also a temporary connection when the device
         // is being paired, without any data transmission

         if (! isConnected) {
            LOGL("Connected")
            Sounds::playConnect();
         }
         isConnected = 1;

         // prevent auto off
         lastActiveMs = timeMs;
         loopConnected();
      }
   }
   else {
      // stop sensor processing when battery switch is off
      // never trigger auto off
      if (isEnabled)
         disable();

      // prevent auto off
      lastActiveMs = timeMs;
      // device inactive except for charging
      loopPassive();
   }
}

/**
 * Configure pins etc
 */
static void configure(void) {

   clearLEDs();

   // Battery Monitoring
   pinMode(PIN_BATTV_ENABLE, OUTPUT);
   digitalWrite(PIN_BATTV_ENABLE, LOW);
   pinMode(PIN_BATTV, INPUT);
   analogReference(ANALOG_REFV);  // 2.4 V reference
   analogReadResolution(12);

   // USB / Charging Monitoring
   // USB Voltage divided, Low <= 0.99V, High >= 2.31V
   pinMode(PIN_USB, INPUT);

   // Charging Rate
   // Slow = HIGH = Charge current 50mA
   // Fast = LOW = Charge current 100mA
   // always slow at beginning
   pinMode(PIN_HICHG, OUTPUT);
   digitalWrite(PIN_HICHG, HIGH);

   // Charging status
   pinMode(PIN_CHG, INPUT);

   // Setup Buzzer LOW = no current
   pinMode(PIN_BUZZER, OUTPUT);
   digitalWrite(PIN_BUZZER, LOW);
}

/**
 * Initial application setup
 */
void appSetup(void) {

   configure();

   // Initialise Serial Monitor
   if (DEBUG) {
      Serial.begin(115200);
      while (! Serial) {
         delay(500);
      }
   }

   // Start I2C
   Wire.begin();

   // Initialise Sensirion SDP
   while (sdp.init() != 0) {
      LOGL("Wait for SDP")
      delay(500);
   }

   // Initialise Sensirion SHT
   sht.begin(Wire);

   // Initialise Infineon DPS sensor
   while (!dps.begin_I2C()) {
      LOGL("Wait for DPS")
      delay(500);
   }

   // Take 64 samples/sec and use average of last 8 samples
   // Temperature compensation is more accurate when measurements
   // are taken together
   dps.configurePressure(DPS310_64HZ, DPS310_8SAMPLES);
   dps.configureTemperature(DPS310_64HZ, DPS310_8SAMPLES);

   // initialise status and start timers
   resetStatus();
   resetTimers();

   // Get initial battery and USB power state
   managePower();

   if (PwrSupply::isBatDrained()) {
      suspend();
      return;
   }

   // Prepare BLE service
   Remote.init(DEVICE_NAME);

   // Set initial on/off status
   if (DEBUG || ACTIVE_WITH_USB || PwrSupply::isBatPowered())
      enable();
   else
      disable();
}
