/**
 *  @file ptvBleInterface.cpp
 *  @brief Extension of BleInterface providing the PTV / PTVSOAR protocol formats
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */
#include "ptvBleInterface.h"

#include <Arduino.h>         // Added by Arduino in build process

#include "debug.h"
#include "bleInterface.h"

//! Optionally append battery voltage to BLE messages
#define SEND_VOLT        0

//! Size of data chunks (bytes) sent via BLE interface
#define CHUNK_SZ         BLE_ATT_LEN

//! Time ms to wait after sending one chunk
#define CHUNK_DELAY_MS   0


/**
 * Must be called periodically to check for received commands
 * @param cmd
 * @return 1 if a new command was received, else 0
 */
int PtvBleInterface::getReceived(String &cmd) {

   poll();

   if (command.length() > 0) {
      cmd = command;
      command = "";
      return 1;
   }
   return 0;
}

/**
 * Periodically called to poll BLE connection for received data chunks / commands.
 * Commands / data chunks must be \n terminated.
 * Chunk length is limited by buffer size.
 */
void PtvBleInterface::poll() {

   // bytewise read the BLE UART Adafruit FIFO,
   // which is hard configured with item size 1
   uint8_t ch = '\0';
   while (bleuart.read(& ch, 1)) {
      if (ch == '\n') {
         // command completed, save it and clear buffer
         command = rcvBuffer;
         clrBuffer();
         break;
      }
      else if (ch != '\r' && ch != '\0') {
         if (pos < sizeof(rcvBuffer)) {
            // concatenate as long as buffer is not full
            rcvBuffer[pos] = ch;
            pos++;
         }
         else {
            // buffer full, still no line end found
            // forcefully complete / split
            command = rcvBuffer;
            clrBuffer();
            break;
         }
      }
   }
   if (! isConnected()) {
      command = "";
      clrBuffer();
   }
}

/**
 * Transmit one message as multiple chunks via BLE
 * @param msg  String with message ready to send including line end
 */
void PtvBleInterface::sendChunked(String & msg) {

   // send data in multiple chunks as required for old BLE versions
   // after each chunk a small delay is needed to avoid overflow of the FIFO buffer

   // append checksum
   byte cksum = 0;
   for (unsigned int i = 1; i < msg.length(); i++) // skip the dollar sign
      cksum ^= (byte) msg[i];  // bitwise exclusive OR
   char nmea[8 + 1];
   sprintf(nmea, "%02x", cksum);

   msg += "*";
   msg += nmea;
   // use LOGL macro without the \n to get proper terminal output
   LOGL(msg)
   msg += "\n";

   unsigned int msglen = msg.length();
   unsigned int chunks = msglen / CHUNK_SZ;
   unsigned int remain = msglen - (chunks * CHUNK_SZ);

   // write the complete chunks
   String part;
   int p = 0;
   for (unsigned int i = 0; i < chunks; i++) {
      part = msg.substring(p, p + CHUNK_SZ);
      p += CHUNK_SZ;

      sendData(part.c_str(), part.length());
      delay(CHUNK_DELAY_MS);
   }

   // write the remainder
   if (remain > 0) {
      part = msg.substring(p, p + remain);

      sendData(part.c_str(), part.length());
      delay(CHUNK_DELAY_MS);
   }
}

/**
 * Send a PTVSOAR meta data message
 * @param mna Manufacturer name
 * @param mmo Device model
 * @param msn Sevice serial number
 */
void PtvBleInterface::sendMetaMsg(const String &mna, const String &mmo, const String &msn) {

   String msg = "$PTVSOAR";
   msg += ",MNA," + mna;
   msg += ",MMO," + mmo;
   msg += ",MSN," + msn;

   sendChunked(msg);
}

/**
 * Send a message using the long PTVSOAR message format
 *
 * NOTE: For devices with BLE speed limitations this format is not useful
 * @param data  sensor data record to be sent
 */
void PtvBleInterface::sendLongMsg(const SensorData &data) {
   // $PTVSOAR,<type>,<value>,<type>,<value>,...*<checksum>\n
   // 11111111111111111111222222222222222222223333333333333333333344444444444444444444555555555555555555556666666666  chunk
   // 0         0         0         0         0         0         0         0         0         0         1           byte index
   // 0         1         2         3         4         5         6         7         8         9         0           byte index
   // 01234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789  byte index
   // $PTVSOAR,MNA,Flying Silicon,MMO,FS-SP4,MSN,23-001,PIT,-00.02,PRS,0987.654,OAT,+24.7,OAH,058,PCT,100,CHG,1*FF\n
   // $PTVSOAR,PIT,-00.02,PRS,0987.654,OAT,+24.7,OAH,058,PCT,100,CHG,1*FF\n

   String msg;

   msg = "$PTVSOAR";
   // optionally the full meta information can be sent
   // msg = META_HEADER;

   char buf[20];

   msg += ",PIT,";
   sprintf(buf, "%06.2f", data.diffPa); // Pitot Differential Pressure 999.99
   msg += buf;

   msg += ",PRS,";
   sprintf(buf, "%08.3f", data.absHpa); // Atmospheric Pressure 9999.999
   msg += buf;

   msg += ",OAT,";
   sprintf(buf, "%+05.1f", data.tempC); // Air Temperature -99.0 .. +99.0
   msg += buf;

   msg += ",OAH,";
   sprintf(buf, "%03.0f", data.humPct); // Air Relative Humidity 000 - 100
   msg += buf;

   msg += ",PCT,";
   sprintf(buf, "%03.0f", data.batPct); // Battery State of Charge % 000 - 100
   msg += buf;

   msg += ",CHG,";
   sprintf(buf, "%d", data.isCharging); // Battery Charging Flag
   msg += buf;

   if (SEND_VOLT) {
      msg += ",VOL,";
      sprintf(buf, "%4.2f", data.batVolt);
      msg += buf;
   }

   // append checksum
   byte cksum = 0;
   for (unsigned int i = 1; i < msg.length(); i++) // skip the dollar sign
      cksum ^= (byte) msg.charAt(i);  // bitwise exclusive OR
   char nmea[8 + 1];
   sprintf(nmea, "%02x", cksum);

   msg += "*";
   msg += nmea;

   // use LOGL macro without the \n to get proper terminal output
   LOGL(msg)

   msg += "\n";
   sendChunked(msg);
}

/**
 * Send a message using the short PTV format
 *
 * Length tailored to 2x20 bytes for BLE 4.1 and earlier
 * @param data  sensor data record to be sent
 */
void PtvBleInterface::sendShortMsg(const SensorData &data) {
   // $PTV,<pitotPa>,<baroHpa>,<tempC>,<humidPct>,<battPct>,<battChg>*<checksum>\n
   // 1111111111111111111122222222222222222222  chunk
   // 0         1         2         3           byte index
   // 0123456789012345678901234567890123456789  byte index
   // $PTV,000.00,0987.105,+20.5,058,100,1*FF\n

   String msg = "$PTV";

   char buf[100]; // enough
   sprintf(buf,
         ",%06.2f"
         ",%08.3f"
         ",%+05.1f"
         ",%03.0f"
         ",%03.0f"
         ",%d"
         , data.diffPa // Pitot Differential Pressure 999.99
         , data.absHpa // Atmospheric Pressure 9999.999
         , data.tempC // Air Temperature -99.9 .. +99.9
         , data.humPct // Air Relative Humidity 000 - 100
         , data.batPct  // Battery State of Charge % 000 - 100
         , data.isCharging // Battery Charging Flag
   );
   msg += buf;

   if (SEND_VOLT) {
      sprintf(buf, ",%4.2f", data.batVolt);
      msg += buf;
   }

   sendChunked(msg);
}

