/**
 *  @file ptvBleInterface.h
 *  @brief Extension of BleInterface providing the PTV / PTVSOAR protocol formats
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */
#ifndef _PTVBLEINTERFACE_H_
#define _PTVBLEINTERFACE_H_

#include "bleInterface.h"

/**
 * Interface for communication via BLE UART
 */
class PtvBleInterface : public BleInterface {

public:

   //! Sensor data structure
   typedef struct {
      //! SDP31 differential pressure Pa
      float diffPa = 0.0;
      //! DPS310 barometric pressure hPa
      float absHpa = 0.0;
      //! SHT temperature °C
      float tempC = 0.0;
      //! SHT humidity %
      float humPct = 0.0;

      //! Battery Voltage
      float batVolt = 0.0;
      //! Battery charge %
      float batPct = 0.0;
      //! Charging state
      int isCharging = 0;

   } SensorData;

   //! Default c-tor
   PtvBleInterface() { clrBuffer(); }

   void sendMetaMsg(const String &mna, const String &mmo, const String &msn);
   void sendShortMsg(const SensorData &data);
   void sendLongMsg(const SensorData &data);
   void sendChunked(String & msg);

   int getReceived(String &cmd);
   //! Clear anything received so far
   void clrReceived() { clrBuffer(); command = ""; }

protected:

   void poll(); // no longer needs to be public

   //! Clear internal receive buffer
   void clrBuffer() { memset(rcvBuffer, 0x00, sizeof(rcvBuffer)); pos = 0; }
   //! Holds last received text line / command
   String command;
   //! Receive data buffer, same size as UART FIFO
   char rcvBuffer[FIFO_SZ];
   //! Write position for next byte
   unsigned int pos;
};


#endif // _PTVBLEINTERFACE_H_
