/**
 *  @file pwrSupply.cpp
 *  @brief Manage power supply status of the device
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */
#include <Arduino.h>

#include "pwrSupply.h"
#include "pindefs.h"
#include "debug.h"

//! Min voltage limit to protect battery
#define MinBattV     3.6

//! Enable fast charging (100mA)
#define FAST_CHARGE  1


namespace PwrSupply {

typedef enum {
   //! Powered by USB only
   PwrUsbOnly     = 0,
   //! Powered by battery only
   PwrBattOnly    = 1,
   //! Powered by battery and connected to USB
   PwrConnected   = 2,
   //! Battery drained
   PwrBattDrained = 3,
} Supply;

//! Power supply status
static int pwrStat = PwrUsbOnly;
//! Battery Voltage
static float batVolt = 0.0;
//! Battery charge %
static float batPct = 0.0;
//! Charging state
static int charging = 0;

/**
 * Check the power status
 * @return 1 if charging, else 0
 */
int isCharging(void) {
   return charging;
}
/**
 * Check the power status
 * @return Last estimated remaining battery percent
 */
float getBatPct(void) {
   return batPct;
}

/**
 * Check the power status
 * @return Last measured battery voltage (depends on charging state)
 */
float getBatVolt(void) {
   return batVolt;
}

/**
 * Check the power status
 * @return 1 if USB power is connected, else 0
 */
int isUsbConnected(void) {
   return (pwrStat == PwrConnected) ? 1 : 0;
}

/**
 * Check the power status
 * @return 1 if the device is powered by USB only, else 0
 */
int isUsbOnly(void) {
  return (pwrStat == PwrUsbOnly) ? 1 : 0;
}

/**
 * Check the power status
 * @return 1 if the device is powered by battery without USB charging, else 0
 */
int isBatPowered(void) {
   return (pwrStat == PwrBattOnly) ? 1 : 0;
}

/**
 * Check the power status
 * @return 1 if the battery appears to be drained, else 0
 */
int isBatDrained(void) {
   return (pwrStat == PwrBattDrained) ? 1 : 0;
}

/**
 * Clear the power status
 */
void reset(void) {
   pwrStat = PwrUsbOnly;
}

/**
 * Update the power status
 */
void updateState(void) {

   // Battery sense pin
   // BQ25100 circuit behaviour is not quite clear without battery:
   // at 100mA: voltage toggles 3.7 - 4.1
   // at 50mA: voltage low

   // CHG pin logic
   // Bat  Usb  Chg
   // off  on   goes on / 1-3x when plugged in, then off
   // on   off  off
   // on   on   goes on when plugged in, then off if batt is full

   // USB voltage Low <= 0.99V, High >= 2.31V
   const int usbOn = digitalRead(PIN_USB);

   // Charging indicator
   charging = digitalRead(PIN_CHG) ? 0 : 1;

   // Enable fast charge
   // only if the chg pin indicates that charging is active == battery connected
   if (FAST_CHARGE && usbOn && charging)
      digitalWrite(PIN_HICHG, LOW); // on
   else
      digitalWrite(PIN_HICHG, HIGH); // off

   // Battery level
   // at 100mA this varies 3.7 - 4.1 when battery disconnected !
   const int adcRead = analogRead(PIN_BATTV);
   const float adcV = 2.4 * adcRead / 4096.0 * 1510.0 / 510.0;

   if (usbOn) {
      // USB connected
      // no reliable percentage known
      if (adcV < 3.0 && ! charging) {
         // running without battery
         batPct = 100.0;
         batVolt = 0;
         pwrStat = PwrUsbOnly;
      }
      else if (adcV < 3.6) {
         // battery voltage low
         batPct = 0.0;
         batVolt = adcV;
         pwrStat = PwrConnected;
      }
      else {
         // with / without battery, power connnected
         batPct = 100;
         batVolt = adcV;
         pwrStat = PwrConnected;
      }
   }
   else {
      // USB not connected
      if (adcV < MinBattV) {
         // for sure empty
         batPct = 0.0;
         batVolt = adcV;
         pwrStat = PwrBattDrained;
      }
      else if (adcV <= 4.2) {
         // formula applicable
         float levelPct = -4370.0 + (2105.0 * adcV) - (247.5144 * adcV * adcV);
         // limit range to 0,100 !
         levelPct = max(0.0, min(100.0, levelPct));

         if (pwrStat != PwrBattOnly) {
            // initialise percentage when starting to discharge
            batPct = levelPct;
         }
         else if (batPct <= 0 || abs(batPct - levelPct) > 20.0) {
            // big step (should not happen)
            batPct = levelPct;
         }
         else {
            // IIR filter to hide spurious small changes
            batPct = batPct + (levelPct - batPct) * 0.05;
         }
         batVolt = adcV;
         pwrStat = PwrBattOnly;
      }
      else {
         // full
         batPct = 100.0;
         batVolt = adcV;
         pwrStat = PwrBattOnly;
      }
   }

   if (DEBUG) {
      LOG("Power: ")
      switch (pwrStat) {
         case PwrUsbOnly: LOGL("USB Only")
            break;
         case PwrBattOnly: LOGL("Battery Only")
            break;
         case PwrConnected: LOGL("USB Connected")
            break;
         case PwrBattDrained: LOGL("Battery Drained")
            break;
         default: LOGL("?")
            break;
      }
      LOG("Adc: ")
      LOG(adcRead)
      LOG("  ")
      LOG(batVolt)
      LOG(" V")
      LOG("  ")
      LOG(batPct)
      LOGL(" %")
   }
}


}
