/**
 *  @file sounds.cpp
 *  @brief Varios methods to produce the buzzer sounds
 *
 *  Arduino code for Flying Silicon Airspeed/Vario Probe and TotalVario Android app.
 *
 *  For more information refer to:
 *  https://flyingsilicon.blogspot.com/ <br>
 *  https://totalvario.blogspot.com/
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  See the GNU General Public License <http://www.gnu.org/licenses/>
 *  for more details.
 */

// Utility functions for buzzer sounds

#include "pindefs.h"
#include "sounds.h"
#include <Arduino.h>


namespace Sounds {

//! Buzzer pin to be used
static int buzzerPin = 0;

//! Duration of a whole note
static const int WholeNote = (60000 * 4) / 90;


/**
 * Enable/disable all sounds on specified buzzer
 * @param  pin  The pin where the buzzer is connected, or 0 to disable sound 
 */
void setBuzzer(int pin) {

   buzzerPin = pin;
}

/**
 * Activate buzzer tone and return immediately
 * Caution: Can leave the buzzer transistor open and drain the battery very fast.
 * @param freq  Tone frequency Hz
 * @param ms  Tone duration ms
 */
static void sound(int freq, int ms) {

   if (buzzerPin <= 0)
      return;

   if (freq < BUZZER_MIN_HZ)
      freq = BUZZER_MIN_HZ;
   else if (freq > BUZZER_MAX_HZ)
      freq = BUZZER_MAX_HZ;

   if (ms < BUZZER_MIN_MS)
      ms = BUZZER_MIN_MS;

   tone(buzzerPin, freq, ms);
}

/**
 * Stop sound
 */
static void silence(void) {

   if (buzzerPin <= 0)
      return;

   noTone(buzzerPin);
   digitalWrite(buzzerPin, LOW);
}

/**
 * Let the buzzer and *be sure to switch it properly off afterwards*.
 * @param freq  Tone frequency Hz
 * @param ms  Tone duration ms
 */
void beep(int freq, int ms) {

   if (buzzerPin <= 0)
      return;

   sound(freq, ms);
   delay(ms);
   silence();
}

/**
 * Play a melody on the buzzer
 *
 * The argument must contain a sequence of int pairs (freq, ms),
 * terminated with MELODY_END
 * @param melody  Melody definition array
 */
void playMelody(const int melody[]) {

   int noteDuration = 0;
   int i = 0;
   while (melody[i] >= BUZZER_MIN_HZ) {

      const int divider = melody[i + 1];

      if (divider == 0) {
         continue;
      }
      else if (divider > 0) {
         // regular note, just proceed
         noteDuration = WholeNote / divider;
      }
      else if (divider < 0) {
         // dotted notes are represented with negative durations!
         noteDuration = WholeNote / abs(divider);
         noteDuration *= 1.5;  // increases the duration in half for dotted notes
      }

      beep(melody[i], noteDuration * 0.9);

      i += 2;
   }
}

/**
 * Sound for battery level when USB is connected
 */
void playCharging(void) {

   beep(NOTE_A1, 100);
   delay(100);
   beep(NOTE_A2, 150);
}

/**
 * Sound indicating the battery status
 * @param batPct  remaining capacity percent
 */
void playBattery(float batPct) {

   const int PauseMs = 200;
   const int ToneMs = 150;

   int n; // number of beep repetitions
   if (batPct > 90) {
      // fully charged, 5 beeps, 5 bars in display
      for (n = 0; n < 5; n++) {
         beep(NOTE_A2, ToneMs);
         delay(PauseMs);
      }
   }
   else if (batPct > 78) {
      // ~80% left
      for (n = 0; n < 4; n++) {
         beep(NOTE_A2, ToneMs);
         delay(PauseMs);
      }
   }
   else if (batPct > 58) {
      // ~60% left
      for (n = 0; n < 3; n++) {
         beep(NOTE_A2, ToneMs);
         delay(PauseMs);
      }
   }
   else if (batPct > 38) {
      // ~40% left
      for (n = 0; n < 2; n++) {
         beep(NOTE_A2, ToneMs);
         delay(PauseMs);
      }
   }
   else if (batPct > 18) {
      // ~20% left
      for (n = 0; n < 1; n++) {
         beep(NOTE_A2, ToneMs);
         delay(PauseMs);
      }
   }
   else {
      // discharged, 1 low beep
      beep(NOTE_A1, 250);
      delay(PauseMs);
   }
}

/**
 * Sound when BLE central is connected
 */
void playConnect(void) {

   beep(NOTE_G3, 100);
   delay(100);
   beep(NOTE_G3, 100);
}

/**
 * Sound when BLE central is disconnected
 */
void playDisconnect(void) {

   beep(NOTE_A2, 100);
   delay(100);
   beep(NOTE_A2, 100);
   delay(100);
   beep(NOTE_A2, 100);
}

/**
 * Sound when device goes inactive
 */
void playSwitchOff(void) {

   beep(NOTE_A1, 500);
}

} // namespace Sounds
